"""Main parsing module."""
from __future__ import annotations
import re
import typing
from collections.abc import KeysView

from . import settings, structs


_RE_FLAGS: re.RegexFlag = re.I | re.M | re.S
TITLE_TAG_RE: typing.Final[re.Pattern] = re.compile(r"<title>\s*(.*?)\s*</title>", flags=_RE_FLAGS)
META_TAGS_RE: typing.Final[re.Pattern] = re.compile(r"<meta([^>]*)>", flags=_RE_FLAGS)
TAG_ATTRS_RE: typing.Final[re.Pattern] = re.compile(
    r"(?:([^\s=\"']+)\s*=\s*(?:(?:[\"'](.*?)[\"'])|[^\s\"']*))", flags=_RE_FLAGS
)


def _extract_social_tags_from_precusor(
    all_tech_attrs: list[dict[str, structs.ValuesGroup]],
    media_type: typing.Literal[structs.WhatToParse.OPEN_GRAPH, structs.WhatToParse.TWITTER],
) -> list[structs.OneMetaTag]:
    possible_settings_for_parsing: dict[str, typing.Union[str, tuple]] = settings.SETTINGS_FOR_SOCIAL_MEDIA[media_type]
    output_buffer: list[structs.OneMetaTag] = []
    for one_attr_group in all_tech_attrs:
        og_tag_name: str = ""
        tech_keys: KeysView[str] = one_attr_group.keys()
        for attr_name in possible_settings_for_parsing["prop"]:
            if attr_name in tech_keys and one_attr_group[attr_name].normalized.startswith(
                possible_settings_for_parsing["prefix"]
            ):
                og_tag_name = one_attr_group[attr_name].normalized.replace(
                    str(possible_settings_for_parsing["prefix"]), ""
                )
            if og_tag_name and "content" in tech_keys and one_attr_group["content"].original:
                output_buffer.append(structs.OneMetaTag(name=og_tag_name, value=one_attr_group["content"].original))
                break
    return output_buffer


def _extract_basic_tags_from_precursor(
    all_tech_attrs: list[dict[str, structs.ValuesGroup]]
) -> list[structs.OneMetaTag]:
    output_buffer: list[structs.OneMetaTag] = []
    for one_attr_group in all_tech_attrs:
        tech_keys: KeysView[str] = one_attr_group.keys()

        if len(output_buffer) == len(settings.BASIC_META_TAGS):
            break

        for one_ordinary_meta_tag in settings.BASIC_META_TAGS:
            if (
                "name" in tech_keys
                and one_attr_group["name"].normalized == one_ordinary_meta_tag
                and "content" in one_attr_group
                and one_attr_group["content"].original
            ):
                output_buffer.append(
                    structs.OneMetaTag(
                        name=one_ordinary_meta_tag,
                        value=one_attr_group["content"].original,
                    )
                )
    return output_buffer


def _extract_all_other_tags_from_precursor(
    all_tech_attrs: list[dict[str, structs.ValuesGroup]]
) -> list[structs.OneMetaTag]:
    output_buffer: list[structs.OneMetaTag] = []
    for one_attr_group in all_tech_attrs:
        tech_keys: KeysView[str] = one_attr_group.keys()

        should_we_skip: bool = False
        for one_config in settings.SETTINGS_FOR_SOCIAL_MEDIA.values():
            for attr_name in one_config["prop"]:
                if attr_name in tech_keys and one_attr_group[attr_name].normalized.startswith(one_config["prefix"]):
                    should_we_skip = True
                    break
        if should_we_skip:
            continue

        if "name" in tech_keys:
            if one_attr_group["name"].normalized in settings.BASIC_META_TAGS:
                continue
            if "content" in one_attr_group and one_attr_group["content"].original:
                output_buffer.append(
                    structs.OneMetaTag(
                        name=one_attr_group["name"].normalized,
                        value=one_attr_group["content"].original,
                    )
                )
    return output_buffer


def parse_meta_tags_from_source(source_code: str, what_to_parse=settings.DEFAULT_PARSE_GROUP) -> structs.TagsGroup:
    """Parse meta tags from source code."""
    builded_result: structs.TagsGroup = structs.TagsGroup()

    if structs.WhatToParse.TITLE in what_to_parse:
        possible_match: typing.Optional[re.Match] = TITLE_TAG_RE.search(source_code)
        if possible_match:
            possible_groups: tuple[typing.Any, ...] = possible_match.groups()
            if len(possible_groups) > 0:
                builded_result.title = str(possible_groups[0])

    if (
        structs.WhatToParse.OPEN_GRAPH in what_to_parse
        or structs.WhatToParse.TWITTER in what_to_parse
        or structs.WhatToParse.BASIC in what_to_parse
        or structs.WhatToParse.OTHER in what_to_parse
    ):
        raw_tags_attrs: typing.Final[list] = META_TAGS_RE.findall(source_code)
        normalized_meta_attrs: list[dict[str, structs.ValuesGroup]] = []
        for one_raw_attrs_row in raw_tags_attrs:
            prepared_attrs: dict = {}
            for attr_key, attr_value in dict(TAG_ATTRS_RE.findall(one_raw_attrs_row)).items():
                prepared_attrs[attr_key.lower().strip()] = structs.ValuesGroup(attr_value, attr_value.lower().strip())
            normalized_meta_attrs.append(prepared_attrs)
        del raw_tags_attrs

        if structs.WhatToParse.OPEN_GRAPH in what_to_parse:
            builded_result.open_graph = _extract_social_tags_from_precusor(
                normalized_meta_attrs, structs.WhatToParse.OPEN_GRAPH
            )

        if structs.WhatToParse.TWITTER in what_to_parse:
            builded_result.twitter = _extract_social_tags_from_precusor(
                normalized_meta_attrs, structs.WhatToParse.TWITTER
            )

        if structs.WhatToParse.BASIC in what_to_parse:
            builded_result.basic = _extract_basic_tags_from_precursor(normalized_meta_attrs)

        if structs.WhatToParse.OTHER in what_to_parse:
            builded_result.other = _extract_all_other_tags_from_precursor(normalized_meta_attrs)

    return builded_result


__all__ = ["parse_meta_tags_from_source"]



"""Snippets helper functions."""
from .parse import parse_meta_tags_from_source, settings, structs


def parse_snippets_from_source(source_code: str) -> structs.SnippetGroup:
    """Parse snippets from source code.

    This function use «ugly» magic like getattr, setattr.
    This is sad, but i don't find at the bottom of my mind a better solution at the moment.
    If you have a more reasonable solution — feel free to push me, I will be grateful for any advice.
    """
    parsed_data: structs.TagsGroup = parse_meta_tags_from_source(
        source_code, (structs.WhatToParse.OPEN_GRAPH, structs.WhatToParse.TWITTER)
    )
    prepared_result: structs.SnippetGroup = structs.SnippetGroup()
    sm_group: str
    one_tag: structs.OneMetaTag
    for sm_group in ("twitter", "open_graph"):
        new_snippet: structs.SocialMediaSnippet = structs.SocialMediaSnippet()
        for one_tag in getattr(parsed_data, sm_group):
            if one_tag.name in settings.SOCIAL_MEDIA_SNIPPET_WHAT_ATTRS_TO_COPY:
                setattr(new_snippet, one_tag.name.replace(":", "_"), one_tag.value)
        setattr(prepared_result, sm_group, new_snippet)
    return prepared_result


__all__ = ["parse_snippets_from_source"]



"""Package settings."""
from __future__ import annotations
import typing

from . import structs


BASIC_META_TAGS: typing.Final[tuple[str, ...]] = ("title", "description", "keywords", "robots", "viewport")
SETTINGS_FOR_SOCIAL_MEDIA: typing.Final[
    dict[
        typing.Literal[structs.WhatToParse.OPEN_GRAPH, structs.WhatToParse.TWITTER], dict[str, typing.Union[str, tuple]]
    ]
] = {
    structs.WhatToParse.OPEN_GRAPH: {"prop": ("property",), "prefix": "og:"},
    # weird thing about twitter: it use name and property simultaneously
    # i mean name is old format, property is new, but all currently accepted as i see at the moment
    structs.WhatToParse.TWITTER: {"prop": ("name", "property"), "prefix": "twitter:"},
}
DEFAULT_PARSE_GROUP: typing.Final[tuple[int, ...]] = (
    structs.WhatToParse.TITLE,
    structs.WhatToParse.BASIC,
    structs.WhatToParse.OPEN_GRAPH,
    structs.WhatToParse.TWITTER,
    structs.WhatToParse.OTHER,
)
SOCIAL_MEDIA_SNIPPET_WHAT_ATTRS_TO_COPY: typing.Final[tuple[str, ...]] = (
    "title",
    "description",
    "url",
    "image",
    "image:width",
    "image:height",
)



"""Main parsing module."""
from __future__ import annotations
import dataclasses
import enum


@dataclasses.dataclass
class OneMetaTag:
    """Helper public tag wrapper."""

    name: str
    value: str


@dataclasses.dataclass
class ValuesGroup:
    """Helper inner wrapper."""

    original: str
    normalized: str


@dataclasses.dataclass
class TagsGroup:
    """Return struct."""

    title: str = ""
    basic: list[OneMetaTag] = dataclasses.field(default_factory=list)
    open_graph: list[OneMetaTag] = dataclasses.field(default_factory=list)
    twitter: list[OneMetaTag] = dataclasses.field(default_factory=list)
    other: list[OneMetaTag] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class SocialMediaSnippet:
    """Social media snippet group."""

    title: str = ""
    description: str = ""
    image: str = ""
    image_width: int = 0
    image_height: int = 0
    url: str = ""


@dataclasses.dataclass
class SnippetGroup:
    """Groupping for social media."""

    open_graph: SocialMediaSnippet = SocialMediaSnippet()
    twitter: SocialMediaSnippet = SocialMediaSnippet()


class WhatToParse(enum.IntEnum):
    """Enum for parsing configuration."""

    TITLE: int = 0
    BASIC: int = 1
    OPEN_GRAPH: int = 2
    TWITTER: int = 3
    OTHER: int = 4



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['meta_tags_parser']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.21.1,<0.22.0']

setup_kwargs = {
    'name': 'meta-tags-parser',
    'version': '1.2.0',
    'description': 'Fast and modern meta tags parser (og, twitter, title, description, etc) with snippet support',
    'long_description': '# Meta tags parser\n[![Test, lint, publish](https://github.com/xfenix/meta-tags-parser/actions/workflows/main.yml/badge.svg)](https://github.com/xfenix/meta-tags-parser/actions/workflows/main.yml)\n[![PyPI version](https://badge.fury.io/py/meta-tags-parser.svg)](https://badge.fury.io/py/meta-tags-parser)\n[![codecov](https://codecov.io/gh/xfenix/meta-tags-parser/branch/master/graph/badge.svg)](https://codecov.io/gh/xfenix/meta-tags-parser)\n<a href="https://github.com/psf/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n[![Imports: isort](https://img.shields.io/badge/imports-isort-%231674b1?style=flat&labelColor=ef8336)](https://timothycrosley.github.io/isort/)\n\nFast, modern, pure python meta tags parser and snippet creator with full support of type annotations, py.typed in basic package and structured output. No jelly dicts, only typed structures!  \nIf you want to see what exactly is social media snippets, look at the example:\n![](https://raw.githubusercontent.com/xfenix/meta-tags-parser/master/social-media-snippets.png)\n\n## Requirements\n* Python 3.8+\n* [Httpx](https://www.python-httpx.org/)\n\n## Install\n`pip install meta-tags-parser`\n\n## Usage\n### TL:DR\n1. Parse meta tags from source:\n    ```python\n    from meta_tags_parser import parse_meta_tags_from_source, structs\n\n\n    desired_result: structs.TagsGroup = parse_meta_tags_from_source("""... html source ...""")\n    # desired_result — is what you want\n    ```\n1. Parse meta tags from url:\n    ```python\n    from meta_tags_parser import parse_tags_from_url, parse_tags_from_url_async, structs\n\n\n    desired_result: structs.TagsGroup = parse_tags_from_url("https://xfenix.ru")\n    # and async variant\n    desired_result: structs.TagsGroup = await parse_tags_from_url_async("https://xfenix.ru")\n    # desired_result — is what you want for both cases\n    ```\n1. Parse social media snippet from source:\n    ```python\n    from meta_tags_parser import parse_snippets_from_source, structs\n\n\n    snippet_obj: structs.SnippetGroup = parse_snippets_from_source("""... html source ...""")\n    # snippet_obj — is what you want\n    # access like snippet_obj.open_graph.title, ...\n    ```\n1. Parse social media snippet from url:\n    ```python\n    from meta_tags_parser import parse_snippets_from_url, parse_snippets_from_url_async, structs\n\n\n    snippet_obj: structs.SnippetGroup = parse_snippets_from_url("https://xfenix.ru")\n    # and async variant\n    snippet_obj: structs.SnippetGroup = await parse_snippets_from_url_async("https://xfenix.ru")\n    # snippet_obj — is what you want\n    # access like snippet_obj.open_graph.title, ...\n    ```\n\n**Huge note**: functions `*_from_url` written only for convenience and very error-prone, so any reconnections/error handling — completely on your side.  \nAlso, I don\'t want to add some bloated requirements to achieve robust connections for any users, because they may simply not await any of this from the library. But if you really need this — write me.\n\n### Basic snippets parsing\nLets say you want extract snippet for twitter from html page:\n```python\nfrom meta_tags_parser import parse_snippets_from_source, structs\n\n\nmy_result: structs.TagsGroup = parse_snippets_from_source("""\n    <meta property="og:card" content="summary_large_image">\n    <meta property="og:url" content="https://github.com/">\n    <meta property="og:title" content="Hello, my friend">\n    <meta property="og:description" content="Content here, yehehe">\n    <meta property="twitter:card" content="summary_large_image">\n    <meta property="twitter:url" content="https://github.com/">\n    <meta property="twitter:title" content="Hello, my friend">\n    <meta property="twitter:description" content="Content here, yehehe">\n""")\n\nprint(my_result)\n# What will be printed:\n"""\nSnippetGroup(\n    open_graph=SocialMediaSnippet(\n        title=\'Hello, my friend\',\n        description=\'Content here, yehehe\',\n        image=\'\',\n        url=\'https://github.com/\'\n    ),\n    twitter=SocialMediaSnippet(\n        title=\'Hello, my friend\',\n        description=\'Content here, yehehe\',\n        image=\'\',\n        url=\'https://github.com/\'\n    )\n)\n"""\n# You can access attributes as this\nmy_result.open_graph.title\nmy_result.twitter.image\n# All fields are necessary and will be always available, even if they have not contain data\n# So no need to worry about attributes exsitence (but you may need to check values)\n```\n\n### Basic meta tags parsing\nMain function is `parse_meta_tags_from_source`. It can be used like this:\n```python\nfrom meta_tags_parser import parse_meta_tags_from_source, structs\n\n\nmy_result: structs.TagsGroup = parse_meta_tags_from_source("""... html source ...""")\nprint(my_result)\n\n# What will be printed:\n"""\nstructs.TagsGroup(\n    title="...",\n    twitter=[\n        structs.OneMetaTag(\n            name="title", value="Hello",\n            ...\n        )\n    ],\n    open_graph=[\n        structs.OneMetaTag(\n            name="title", value="Hello",\n            ...\n        )\n    ],\n    basic=[\n        structs.OneMetaTag(\n            name="title", value="Hello",\n            ...\n        )\n    ],\n    other=[\n        structs.OneMetaTag(\n            name="article:name", value="Hello",\n            ...\n        )\n    ]\n)\n"""\n```\nAs you can see from this example, we are not using any jelly dicts, only structured dataclasses. Lets see another example:\n\n```python\nfrom meta_tags_parser import parse_meta_tags_from_source, structs\n\n\nmy_result: structs.TagsGroup = parse_meta_tags_from_source("""\n    <meta property="twitter:card" content="summary_large_image">\n    <meta property="twitter:url" content="https://github.com/">\n    <meta property="twitter:title" content="Hello, my friend">\n    <meta property="twitter:description" content="Content here, yehehe">\n""")\n\nprint(my_result)\n# What will be printed:\n"""\nTagsGroup(\n    title=\'\',\n    basic=[],\n    open_graph=[],\n    twitter=[\n        OneMetaTag(name=\'card\', value=\'summary_large_image\'),\n        OneMetaTag(name=\'url\', value=\'https://github.com/\'),\n        OneMetaTag(name=\'title\', value=\'Hello, my friend\'),\n        OneMetaTag(name=\'description\', value=\'Content here, yehehe\')\n    ],\n    other=[]\n)\n"""\n\nfor one_tag in my_result.twitter:\n    if one_tag.name == "title":\n        print(one_tag.value)\n# What will be printed:\n"""\nHello, my friend\n"""\n```\n\n### If you want to improve speed\nYou can specify what you want to parse:\n```python\nfrom meta_tags_parser import parse_meta_tags_from_source, structs\n\n\nresult: structs.TagsGroup = parse_meta_tags_from_source("""... source ...""",\n    what_to_parse=(WhatToParse.TITLE, WhatToParse.BASIC, WhatToParse.OPEN_GRAPH, WhatToParse.TWITTER, WhatToParse.OTHER)\n)\n```\nIf you reduce this tuple of parsing requirements it may increase overall parsing speed.\n\n## Important notes\n* Any name in meta tag (name or property attribute) will be lowercased\n* I decided to strip `og:` and `twitter:` from original attributes, and let dataclass structures carry this information. If parser met meta tag with property `og:name`, it will be available in `my_result` variable as one element of list `my_result.open_graph`\n* Title of page (e.g. `<title>Something</title>`) will be available as string `my_result.title` (of course, you recieve `Something`)\n* «Standart» tags like title, description (check full list here [./meta_tags_parser/structs.py](./meta_tags_parser/structs.py) in constant `BASIC_META_TAGS`) will be available as list in `my_result.basic`\n* Other tags will be available as list in `my_result.other` attribute, name of tags will be preserved, unlike `og:`/`twitter:` behaviour\n* If you want structured snippets, use `parse_snippets_from_source` function\n\n\n# Changelog\nYou can check https://github.com/xfenix/meta-tags-parser/releases/ release page.\n',
    'author': 'Denis Anikin',
    'author_email': 'ad@xfenix.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/xfenix/meta-tags-parser/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Downloader functions."""
import httpx


def download_page_sync(uri_of_page: str) -> str:
    """Simple yet stupid downloader."""
    return httpx.get(uri_of_page).text


async def download_page_async(uri_of_page: str) -> str:
    """Simple yet stupid async downloader."""
    async with httpx.AsyncClient() as client:
        request_obj: httpx.Response = await client.get(uri_of_page)
        return request_obj.text


__all__ = ["download_page_async", "download_page_sync"]



"""Bunch of public methods."""
from . import download, parse, structs
from .snippets import parse_snippets_from_source


def parse_tags_from_url(web_url: str) -> structs.TagsGroup:
    """Stupid and low quality helper."""
    return parse.parse_meta_tags_from_source(download.download_page_sync(web_url))


async def parse_tags_from_url_async(web_url: str) -> structs.TagsGroup:
    """Stupid and low quality helper."""
    return parse.parse_meta_tags_from_source(await download.download_page_async(web_url))


def parse_snippets_from_url(web_url: str) -> structs.SnippetGroup:
    """Stupid and low quality helper."""
    return parse_snippets_from_source(download.download_page_sync(web_url))


async def parse_snippets_from_url_async(web_url: str) -> structs.SnippetGroup:
    """Stupid and low quality helper."""
    return parse_snippets_from_source(await download.download_page_async(web_url))


__all__ = [
    "parse_tags_from_url",
    "parse_tags_from_url_async",
    "parse_snippets_from_url",
    "parse_snippets_from_url_async",
]
